#! python3  # noqa: E265

"""Helper to use pdal commands"""
# -- IMPORTS

# standard lib
import logging
import re
from pathlib import Path
from re import Match

# 3rd party
from gpf_entrepot_toolbelt.orchestrator.models.bbox import GpfBbox
from gpf_entrepot_toolbelt.orchestrator.status import Status
from gpf_entrepot_toolbelt.utils.geometry_utils import transform_coords_to_4326

# package
from gpf_common_point_cloud.utils.pdal_utils import pdal_info_from_las

# logs
logger = logging.getLogger(__name__)

READER_NAME: str = "readers.las"
MINX_FIELD: str = "minx"
MINY_FIELD: str = "miny"
MAXX_FIELD: str = "maxx"
MAXY_FIELD: str = "maxy"
LAZ_EXTENSION: str = ".laz"

EXTRACT_COORDINATES_REGEX: str = r"([{][^}{]+})"
COORDINATE_REGEX: str = (
    r"[{](?P<field>[a-zA-Z_]+)(/(?P<div>\d+))?(?P<decimal>:(\d+)?([.]\d+)?f)?}"
)


def check_laslaz_file_consistency(
    file: Path,
    expected_srs: str | None = None,
    template: str | None = None,
    origin: dict | None = None,
    tile_size: float | None = None,
) -> tuple[Status, GpfBbox | None]:
    """Check the consistency of the file.

    Args:
        file (Path): path of the file to check
        expected_srs (str): SRS expected
        template (str): Template, example="DALLE_{x_min/1000:04.0f}_{y_max/1000:04.0f}.copc.laz"
        origin (dict): Origin of tiling {"x": string, "y": string}
        tile_size (float): Tile size

    Returns:
        Status: status of the check
    """
    srs: str | None = None
    bbox: GpfBbox | None = None

    # Check the file with PDAL / also check and log Terrascan issue
    metadata: dict = pdal_info_from_las(file)
    if metadata is None:
        return Status.FAILURE, None

    try:
        srs = "EPSG:" + str(metadata[READER_NAME]["srs"]["json"]["id"]["code"])
    except KeyError:
        logger.user_warning(
            f"Aucun SRS n'est défini au sein du fichier. Traitement avec le SRS {expected_srs}"
        )

    if srs or expected_srs:
        coords_4326: list[float] = transform_coords_to_4326(
            [
                metadata[READER_NAME][MINX_FIELD],
                metadata[READER_NAME][MINY_FIELD],
                metadata[READER_NAME][MAXX_FIELD],
                metadata[READER_NAME][MAXY_FIELD],
            ],
            srs if srs else expected_srs,
        )
        if coords_4326 is not None:
            bbox: GpfBbox = GpfBbox(
                tuple[float](
                    [coords_4326[2], coords_4326[0], coords_4326[3], coords_4326[1]]
                )
            )

    # Print the version of the LAS file for debug
    logger.user_info(f"Version du fichier: {_get_file_version(metadata)}")

    # Check the SRS
    if srs and expected_srs and srs != expected_srs:
        logger.user_error(
            f"Le SRS du fichier {srs} n'est pas celui attendu {expected_srs}"
        )
        return Status.FAILURE, bbox

    # Check the Name
    if template is not None:
        name: str = get_name_from_template(metadata, template, origin, tile_size)
        if name != file.name:
            logger.user_error(
                f"Le nom du fichier ({file.name}) n'est pas celui attendu ({name})"
            )
            return Status.FAILURE, bbox

    # Check concordance with tiling
    if origin is not None and tile_size is not None:
        if not _is_file_in_tiling(metadata, origin, tile_size):
            logger.user_error(
                f"Le fichier n'est pas aligné sur la dalle d'origine {origin} et de taille {tile_size}"
            )
            return Status.FAILURE, bbox

    return Status.SUCCESS, bbox


def _get_file_version(metadata: dict) -> str:
    """Get the version of the LAS file.

    Args:
        metadata (dict): metadata of the file

    Returns:
        str: version of the LAS file
    """
    return "{}.{}".format(
        metadata[READER_NAME]["major_version"],
        metadata[READER_NAME]["minor_version"],
    )


def correct_max_coord(coord: float, origin: float, tile_size: float) -> float:
    """
    Correct max coordinate part
    Args:
        coord (float): coordinate part to correct
        origin (float): grid origin
        tile_size (float): tile size

    Returns:
        float: corrected coordinate part
    """

    if tile_size <= 0:
        return coord

    shift = (coord - origin) % tile_size
    if shift != 0:
        return coord - shift + tile_size

    return coord


def correct_min_coord(coord: float, origin: float, tile_size: float) -> float:
    """
    Correct min coordinate part
    Args:
        coord (float): coordinate part to correct
        origin (float): grid origin
        tile_size (float): tile size

    Returns:
        float: corrected coordinate part
    """

    if tile_size <= 0:
        return coord

    shift = (coord - origin) % tile_size
    if shift != 0:
        return coord - shift

    return coord


def is_template_valid(template) -> bool:
    """
    Check if a naming template is valid, which means that it contains at least one X field and one Y field

    Args:
        template (str): Template to check

    Returns:
        bool: True if the template is valid
    """

    coordinates: [str] = re.split(EXTRACT_COORDINATES_REGEX, template)

    found_x_field: bool = False
    found_y_field: bool = False
    for coordinate in coordinates:
        coordinate_fields: Match[str] = re.search(COORDINATE_REGEX, coordinate)

        if coordinate_fields is None:
            continue

        match coordinate_fields.group("field"):
            case "x_min" | "x_max" | "x_mean":
                found_x_field = True
            case "y_min" | "y_max" | "y_mean":
                found_y_field = True

    return found_x_field and found_y_field


def get_name_from_template(
    metadata: dict, template: str, origin: dict, tile_size: float
) -> str | None:
    """
    Compute file name from a template

    Args:
        metadata (dict): Metadata of the file from the PDAL LAS reader
        template (str): Template to use
        origin (dict): Origin of tiling {"x": string, "y": string}
        tile_size (float): Tile size
    Returns:
        str: The computed file name
    """

    reader_metadata: dict = metadata[READER_NAME]
    x_min = correct_min_coord(reader_metadata[MINX_FIELD], origin["x"], tile_size)
    x_max = correct_max_coord(reader_metadata[MAXX_FIELD], origin["x"], tile_size)
    y_min = correct_min_coord(reader_metadata[MINY_FIELD], origin["y"], tile_size)
    y_max = correct_max_coord(reader_metadata[MAXY_FIELD], origin["y"], tile_size)

    values: dict = {
        "x_min": x_min,
        "y_min": y_min,
        "x_max": x_max,
        "y_max": y_max,
        "x_mean": (x_min + x_max) / 2,
        "y_mean": (y_min + y_max) / 2,
    }

    coordinates: [str] = re.split(EXTRACT_COORDINATES_REGEX, template)

    name: str = ""
    for coordinate in coordinates:
        coordinate_fields = re.search(COORDINATE_REGEX, coordinate)

        if coordinate_fields is None:
            name += coordinate

        else:
            computed_coordinate: str = _compute_coordinate(
                values,
                coordinate_fields.group("field"),
                coordinate_fields.group("div"),
                coordinate_fields.group("decimal"),
            )

            if computed_coordinate is None:
                logger.user_error(
                    "Impossible d'utiliser le template pour calculer le nom d'une tuile"
                )
                return None

            name += computed_coordinate

    return name


def _compute_coordinate(
    values: dict, field: str, divider: str | None, decimal_formatter: str | None
) -> str | None:
    """
    Compute a coordinate value from a template part

    Args:
        values (dict): Possible coordinate values
        field (str): The requested field
        divider (str): The divider part of the template
        decimal_formatter (str): The decimal formatter part of the template

    Returns:
        str: The computed coordinate
    """

    # Extract the field value
    if field not in values:
        logger.user_error(f"Le champ {field} n'est pas reconnu par le template")
        return None
    value: float = values[field]

    # Apply divider
    if divider:
        value = value / int(divider)

    # Apply formatting
    if decimal_formatter is None:
        decimal_formatter = ":.0f"
    return ("{" + decimal_formatter + "}").format(value)


def _is_file_in_tiling(metadata: dict, origin: dict, tile_size: float) -> bool:
    """
    Indicate if a las file respect a tiling

    Args:
        metadata (dict): Metadata of the file from the PDAL LAS reader
        origin (dict): Origin X,Y of the tiling
        tile_size (float): Tile size of the tiling

    Returns:
        bool: True if the file respect the tiling, False otherwise
    """

    reader_metadata: dict = metadata[READER_NAME]
    result: bool = True
    if (
        reader_metadata[MAXX_FIELD]
        > correct_min_coord(reader_metadata[MINX_FIELD], origin["x"], tile_size)
        + tile_size
    ):
        logger.user_error(
            f"Les {MINX_FIELD} et {MAXX_FIELD} ne sont pas alignés sur la grille"
        )
        result = False

    if (
        reader_metadata[MAXY_FIELD]
        > correct_min_coord(reader_metadata[MINY_FIELD], origin["y"], tile_size)
        + tile_size
    ):
        logger.user_error(
            f"Les {MINY_FIELD} et {MAXY_FIELD} ne sont pas alignés sur la grille"
        )
        result = False

    return result
