#! python3  # noqa: E265

"""Main module logic."""

# standard lib
import json
import logging
import subprocess
import traceback
from pathlib import Path

# 3rd party
import jsonschema
from gpf_entrepot_toolbelt.orchestrator.status import Status
from gpf_entrepot_toolbelt.utils.geometry_utils import simplify_extent
from jsonschema.exceptions import ValidationError
from shapely import force_2d
from shapely.geometry import MultiPolygon, Polygon
from shapely.geometry.base import BaseGeometry
from shapely.ops import unary_union

# project
from gpf_common_point_cloud.utils.vpc_constants import VPC_FILE_NAME, VPC_SCHEMA

# logs
logger = logging.getLogger(__name__)


def vpc_to_json(vpc_file: Path) -> dict | None:
    """
    Read VPC file into a dict

    Args:
        vpc_file (Path): Path to the vpc file to read

    Returns:
        dict | None: result dict or None if an error occured
    """

    try:
        logger.info(f"Ouverture du fichier {vpc_file}")
        with open(vpc_file) as f:
            vpc: dict = json.load(f)
    except Exception as error:
        logger.user_error(f"Le {VPC_FILE_NAME} n'est pas un JSON lisible'")
        logger.error(
            f"Impossible de lire le fichier {vpc_file} : {error}"
            f"Trace: {traceback.format_exc()}"
        )
        return None

    if "features" not in vpc:
        logger.user_error(f"Le {VPC_FILE_NAME} ne contient pas d'élément 'features'")
        return None

    return vpc


def check_vpc(vpc_file: Path, copc_list: list[str]) -> Status:
    """
    Validate VPC file regarding the structure and the expected copc list

    Args:
        vpc_file (Path): Path to the vpc file to validate
        copc_list (list[str]): List of upload relatives path of expected copc files

    Returns:
        Status: status of the check
    """

    logger.user_info("Vérification du fichier VPC")
    vpc: dict = vpc_to_json(vpc_file)
    if vpc is None:
        return Status.FAILURE

    result: Status = Status.SUCCESS
    for feature in vpc["features"]:
        feature_id: str = feature["id"] if "id" in feature else "UNKNOWN_IDENTIFIER"

        try:
            jsonschema.validate(feature, VPC_SCHEMA)

        except ValidationError as e:
            logger.user_error(
                f"La description JSON de la feature d'identifiant '{feature_id}' du {VPC_FILE_NAME} est incorrecte"
            )
            logger.error(e)
            result = Status.FAILURE
            continue

        href: str = feature["assets"]["data"]["href"]
        if href.__contains__("../"):
            logger.user_error(
                f"La feature d'identifiant '{feature_id}' du {VPC_FILE_NAME} est dans une arboresence montante "
                f"par rapport à {VPC_FILE_NAME} ({href})"
            )
            result = Status.FAILURE

        try:
            copc_list.pop(copc_list.index(str(Path(vpc_file.parent / href).resolve())))

        except ValueError:
            logger.user_error(
                f"La feature d'identifiant '{feature_id}' du {VPC_FILE_NAME} "
                f"n'est pas présente dans la livraison (fichier attendu : {href})"
            )
            result = Status.FAILURE

    if len(copc_list) != 0:
        copc_list_as_str: str = ",".join(copc_list)
        logger.user_error(
            f"Le(s) fichier(s) : '{copc_list_as_str}' ne sont pas présent dans le {VPC_FILE_NAME}"
        )
        result = Status.FAILURE

    return result


def compute_geojson_extent_from_vpc(vpc_file: Path) -> dict | None:
    """
    Compute extent from VPC file

    Args:
        vpc_file: Path, vpc file path

    Returns:
        dict (geojson): global extent simplified
    """

    logger.user_info("Calcul de l'étendue depuis le fichier VPC")
    vpc: dict = vpc_to_json(vpc_file)
    if vpc is None:
        return None

    all_polygons: list[BaseGeometry] = []
    for feature in vpc["features"]:
        try:
            coordinates = feature["geometry"]["coordinates"][0]
            polygon = force_2d(Polygon(coordinates))
            all_polygons.append(polygon)
        except Exception as error:
            logger.error(
                f"Impossible de lire une feature du VPC {vpc_file} : {error}"
                f"Trace: {traceback.format_exc()}"
            )
            return None

    multi_polygon = MultiPolygon(all_polygons)
    union_polygon = unary_union(multi_polygon)

    final_simplified_polygon = simplify_extent(union_polygon, threshold=5000)

    return {"type": "Feature", "geometry": final_simplified_polygon.__geo_interface__}


def build_vpc(input_dir: Path, output_dir: Path) -> Status:
    """
    Generate VPC file from the produced tiling
    output file path will be work_dir / VPC_FILE_NAME (index.vpc)

    Args:
        input_dir (Path): directory containing POINT-CLOUD files
        output_dir (Path): target directory for index.vpc file

    Returns:
        Status: SUCCESS if the job is ok, FAILURE otherwise
    """

    try:
        logger.user_info("Génération du VPC")

        inputs: str = " ".join([str(file.resolve()) for file in input_dir.glob("*")])
        command: str = (
            f"pdal_wrench build_vpc --output={output_dir / VPC_FILE_NAME} {inputs}"
        )
        logger.info("Lancement de la commande pdal_wrench build_vpc")
        logger.debug(f"Détails de la commande : {command}")
        run_result = subprocess.run(command, shell=True, capture_output=True, text=True)

        if run_result.returncode != 0:
            logger.error(f"Code retour : {run_result.returncode}")
            logger.error(f"Stdout : {run_result.stdout}")
            logger.error(f"Stderr : {run_result.stderr}")
            logger.user_error(
                f"Erreur lors de la génération du fichier {VPC_FILE_NAME}"
            )
            return Status.TECHNICAL_ERROR

        return Status.SUCCESS

    except Exception as error:
        logger.user_error(f"Erreur lors de la génération du fichier {VPC_FILE_NAME}")
        logger.error(error, stack_info=True)
        return Status.TECHNICAL_ERROR
