#! python3  # noqa E265
import subprocess

# standard library
import unittest
from pathlib import Path
from unittest.mock import patch

# 3rd party
from gpf_entrepot_toolbelt.orchestrator.status import Status
from gpf_entrepot_toolbelt.utils.gpf_logger import gpf_logger_script

# project
from gpf_common_point_cloud.enums.entwine_enums import EntwineStep
from gpf_common_point_cloud.utils.entwine_utils import (
    determine_subsets_number,
    run_entwine,
)
from gpf_common_point_cloud.utils.laslaz_utils import READER_NAME

# ################################
# ########## Classes #############
# ################################

logger = gpf_logger_script(verbosity=0, title="TestEntwineUtils")


class TestEntwineUtils(unittest.TestCase):
    """Test for entwine_utils functions"""

    FAKE_LASLAZ_UPLOAD_PATH = Path("tests/fixtures/entwine/fake_laslaz_upload")

    TILE_0_0_N: dict = {
        READER_NAME: {"minx": 0, "miny": 0, "maxx": 1, "maxy": 1, "count": 1000000}
    }
    TILE_1_0_N: dict = {
        READER_NAME: {"minx": 1, "miny": 0, "maxx": 2, "maxy": 1, "count": 1000000}
    }
    TILE_0_1_N: dict = {
        READER_NAME: {"minx": 0, "miny": 1, "maxx": 1, "maxy": 2, "count": 1000000}
    }
    TILE_1_1_N: dict = {
        READER_NAME: {"minx": 1, "miny": 1, "maxx": 2, "maxy": 2, "count": 1000000}
    }
    TILE_14_14_N: dict = {
        READER_NAME: {"minx": 14, "miny": 14, "maxx": 15, "maxy": 15, "count": 1000000}
    }
    TILE_15_14_N: dict = {
        READER_NAME: {"minx": 15, "miny": 14, "maxx": 16, "maxy": 15, "count": 1000000}
    }
    TILE_14_15_N: dict = {
        READER_NAME: {"minx": 14, "miny": 15, "maxx": 15, "maxy": 16, "count": 1000000}
    }
    TILE_15_15_N: dict = {
        READER_NAME: {"minx": 15, "miny": 15, "maxx": 16, "maxy": 16, "count": 1000000}
    }
    TILE_0_2_N: dict = {
        READER_NAME: {"minx": 0, "miny": 2, "maxx": 1, "maxy": 3, "count": 1000000}
    }
    TILE_1_2_N: dict = {
        READER_NAME: {"minx": 1, "miny": 2, "maxx": 2, "maxy": 3, "count": 1000000}
    }
    TILE_254_255_N: dict = {
        READER_NAME: {
            "minx": 254,
            "miny": 255,
            "maxx": 255,
            "maxy": 256,
            "count": 1000000,
        }
    }
    TILE_255_255_N: dict = {
        READER_NAME: {
            "minx": 255,
            "miny": 255,
            "maxx": 256,
            "maxy": 256,
            "count": 1000000,
        }
    }
    TILE_05_05_N: dict = {
        READER_NAME: {
            "minx": 0.5,
            "miny": 0.5,
            "maxx": 1.5,
            "maxy": 1.5,
            "count": 1000000,
        }
    }
    TILE_0_0_VS: dict = {
        READER_NAME: {"minx": 0, "miny": 0, "maxx": 1, "maxy": 1, "count": 50000}
    }
    TILE_1_0_VS: dict = {
        READER_NAME: {"minx": 1, "miny": 0, "maxx": 2, "maxy": 1, "count": 50000}
    }
    TILE_0_1_VS: dict = {
        READER_NAME: {"minx": 0, "miny": 1, "maxx": 1, "maxy": 2, "count": 50000}
    }
    TILE_1_1_VS: dict = {
        READER_NAME: {"minx": 1, "miny": 1, "maxx": 2, "maxy": 2, "count": 50000}
    }
    TILE_0_2_VS: dict = {
        READER_NAME: {"minx": 0, "miny": 2, "maxx": 1, "maxy": 3, "count": 50000}
    }
    TILE_1_2_VS: dict = {
        READER_NAME: {"minx": 1, "miny": 2, "maxx": 2, "maxy": 3, "count": 50000}
    }
    TILE_254_255_S: dict = {
        READER_NAME: {
            "minx": 254,
            "miny": 255,
            "maxx": 255,
            "maxy": 256,
            "count": 200000,
        }
    }
    TILE_255_255_S: dict = {
        READER_NAME: {
            "minx": 255,
            "miny": 255,
            "maxx": 256,
            "maxy": 256,
            "count": 200000,
        }
    }
    TILE_LEVEL_0_TEST: dict = {
        READER_NAME: {
            "minx": 0,
            "miny": 0,
            "maxx": 1,
            "maxy": 1,
            "count": 1000,
        }
    }

    def test_run_entwine_ok(self):
        # Given
        config_json_path = Path("tests/fixtures/delivery/OK/generation/config.json")
        with patch("subprocess.run") as run_mock:
            run_mock.return_value = subprocess.CompletedProcess(args="", returncode=0)

            # When
            result: Status = run_entwine(config_json_path, EntwineStep.BUILD)

            # Then
            self.assertEqual(result, Status.SUCCESS)
            run_mock.assert_called_with(
                "entwine build -c tests/fixtures/delivery/OK/generation/config.json",
                shell=True,
                capture_output=True,
                text=True,
            )

    def test_run_entwine_ko(self):
        # Given
        config_json_path = Path("tests/fixtures/delivery/OK/generation/config.json")
        with patch("subprocess.run") as run_mock:
            run_mock.return_value = subprocess.CompletedProcess(
                args="", returncode=1, stdout="STDOUT", stderr="STDERR"
            )

            # When
            result: Status = run_entwine(config_json_path, EntwineStep.BUILD)

            # Then
            self.assertEqual(result, Status.FAILURE)
            run_mock.assert_called_with(
                "entwine build -c tests/fixtures/delivery/OK/generation/config.json",
                shell=True,
                capture_output=True,
                text=True,
            )

    def test_determine_subsets_number_ok_max_tile_4(self):
        # Given
        with patch(
            "gpf_common_point_cloud.utils.entwine_utils.pdal_info_from_las",
            side_effect=[
                self.TILE_0_0_N,
                self.TILE_1_0_N,
                self.TILE_0_1_N,
                self.TILE_1_1_N,
                self.TILE_14_14_N,
                self.TILE_15_14_N,
                self.TILE_14_15_N,
                self.TILE_15_15_N,
            ],
        ):
            # When
            result: int = determine_subsets_number(self.FAKE_LASLAZ_UPLOAD_PATH)

            # Then
            self.assertEqual(result, 4)

    def test_determine_subsets_number_ok_max_tile_16384(self):
        # Given
        with patch(
            "gpf_common_point_cloud.utils.entwine_utils.pdal_info_from_las",
            side_effect=[
                self.TILE_0_0_N,
                self.TILE_1_0_N,
                self.TILE_0_1_N,
                self.TILE_1_1_N,
                self.TILE_0_2_N,
                self.TILE_1_2_N,
                self.TILE_254_255_N,
                self.TILE_255_255_N,
            ],
        ):
            # When
            result: int = determine_subsets_number(self.FAKE_LASLAZ_UPLOAD_PATH)

            # Then
            self.assertEqual(result, 4)

    def test_determine_subsets_number_ok_max_count_4(self):
        # Given
        with patch(
            "gpf_common_point_cloud.utils.entwine_utils.pdal_info_from_las",
            side_effect=[
                self.TILE_0_0_VS,
                self.TILE_1_0_VS,
                self.TILE_0_1_VS,
                self.TILE_1_1_VS,
                self.TILE_0_2_VS,
                self.TILE_1_2_VS,
                self.TILE_254_255_S,
                self.TILE_255_255_S,
            ],
        ):
            # When
            result: int = determine_subsets_number(self.FAKE_LASLAZ_UPLOAD_PATH)

            # Then
            self.assertEqual(result, 4)

    def test_determine_subsets_number_ok_max_subsets(self):
        # Given
        with patch(
            "gpf_common_point_cloud.utils.entwine_utils.pdal_info_from_las",
            side_effect=[
                self.TILE_0_0_N,
                self.TILE_1_0_N,
                self.TILE_0_1_N,
                self.TILE_1_1_N,
                self.TILE_05_05_N,
                self.TILE_15_14_N,
                self.TILE_15_15_N,
                self.TILE_255_255_N,
            ],
        ):
            # When
            result: int = determine_subsets_number(self.FAKE_LASLAZ_UPLOAD_PATH)

            # Then
            self.assertEqual(result, 4)

    def test_determine_subsets_number_ok_min_level(self):
        # Given
        with patch(
            "gpf_common_point_cloud.utils.entwine_utils.pdal_info_from_las",
            side_effect=[
                self.TILE_LEVEL_0_TEST,
                self.TILE_LEVEL_0_TEST,
                self.TILE_LEVEL_0_TEST,
                self.TILE_LEVEL_0_TEST,
                self.TILE_LEVEL_0_TEST,
                self.TILE_LEVEL_0_TEST,
                self.TILE_LEVEL_0_TEST,
                self.TILE_LEVEL_0_TEST,
            ],
        ):
            # When
            result: int = determine_subsets_number(self.FAKE_LASLAZ_UPLOAD_PATH)

            # Then
            self.assertEqual(result, 4)

    def test_determine_subsets_number_nok_pdal_error(self):
        # Given
        with patch(
            "gpf_common_point_cloud.utils.entwine_utils.pdal_info_from_las",
            side_effect=[
                self.TILE_0_0_N,
                self.TILE_1_0_N,
                None,
            ],
        ):
            # When
            result: int = determine_subsets_number(self.FAKE_LASLAZ_UPLOAD_PATH)

            # Then
            self.assertEqual(result, -1)


# ################################
# ####### Stand-alone run ########
# ################################

if __name__ == "__main__":
    unittest.main()
