#! python3  # noqa E265

# standard library
import unittest
from pathlib import Path
from unittest.mock import MagicMock, call, patch

# 3rd party
from gpf_entrepot_toolbelt.orchestrator.models import GpfExtent
from gpf_entrepot_toolbelt.orchestrator.status import Status
from gpf_entrepot_toolbelt.utils.gpf_logger import gpf_logger_script

# project
import gpf_common_point_cloud
from gpf_common_point_cloud.utils.ept_utils import (
    check_entwine_info,
    check_ept_head,
    check_ept_hierarchy,
    clean_pyramid_temporary_files,
    get_extent,
    get_srs,
)

logger = gpf_logger_script(verbosity=0, title="TestEptUtils")


# ################################
# ########## Classes #############
# ################################


class TestEptUtils(unittest.TestCase):
    """Test for core functions"""

    EPT_SRS: str = "EPSG:26913"
    EPT_DIR: str = "tests/fixtures/ept/valid_ept"
    EPT_HEAD: str = EPT_DIR + "/ept.json"
    WORK_DIR_PATH: Path = Path("/tmp")

    def test_check_entwine_info_ok(self):
        # When
        result_status = check_entwine_info(
            self.WORK_DIR_PATH, Path(self.EPT_HEAD), self.EPT_SRS
        )

        # Then
        self.assertEqual(result_status, Status.SUCCESS)

    def test_check_check_ept_hierarchy_ok(self):
        # When
        result_status = check_ept_hierarchy(Path(self.EPT_DIR))

        # Then
        self.assertEqual(result_status, Status.SUCCESS)

    def test_check_check_ept_hierarchy_nok_missing_files(self):
        # Given
        upload_dir_path: Path = Path(
            "tests/fixtures/ept/invalid_ept/missing_files_in_hierarchy"
        )

        # When
        result_status = check_ept_hierarchy(upload_dir_path)

        # Then
        self.assertEqual(result_status, Status.FAILURE)

    def test_check_check_ept_hierarchy_nok_too_much_files(self):
        # Given
        upload_dir_path: Path = Path(
            "tests/fixtures/ept/invalid_ept/too_much_files_in_hierarchy"
        )

        # When
        result_status = check_ept_hierarchy(upload_dir_path)

        # Then
        self.assertEqual(result_status, Status.FAILURE)

    def test_check_entwine_info_nok_entwine_invalid(self):
        # Given
        ept_head = Path("tests/fixtures/ept/invalid_ept/no_data_ept/ept.json")

        # When
        result_status = check_entwine_info(self.WORK_DIR_PATH, ept_head, self.EPT_SRS)

        # Then
        self.assertEqual(result_status, Status.FAILURE)

    def test_check_entwine_info_nok_bad_srs(self):
        # When
        result_status = check_entwine_info(
            self.WORK_DIR_PATH, Path(self.EPT_HEAD), "EPSG:2154"
        )

        # Then
        self.assertEqual(result_status, Status.FAILURE)

    def test_check_ept_head_ok(self):
        # Given
        gpf_common_point_cloud.utils.ept_utils.check_entwine_info = MagicMock(
            return_value=Status.SUCCESS
        )

        # When
        result_status, result_ept_head = check_ept_head(
            self.WORK_DIR_PATH, Path(self.EPT_DIR), self.EPT_SRS
        )

        # Then
        self.assertEqual(result_status, Status.SUCCESS)
        self.assertEqual(result_ept_head, Path(self.EPT_HEAD))

        gpf_common_point_cloud.utils.ept_utils.check_entwine_info.assert_called_with(
            self.WORK_DIR_PATH, Path(self.EPT_HEAD), expected_srs=self.EPT_SRS
        )

    def test_check_ept_head_nok_error_entwine_info(self):
        # Given
        gpf_common_point_cloud.utils.ept_utils.check_entwine_info = MagicMock(
            return_value=Status.TECHNICAL_ERROR
        )

        # When
        result_status, result_ept_head = check_ept_head(
            self.WORK_DIR_PATH, Path(self.EPT_DIR), self.EPT_SRS
        )

        # Then
        self.assertEqual(result_status, Status.TECHNICAL_ERROR)
        self.assertEqual(result_ept_head, None)

        gpf_common_point_cloud.utils.ept_utils.check_entwine_info.assert_called_with(
            self.WORK_DIR_PATH, Path(self.EPT_HEAD), expected_srs=self.EPT_SRS
        )

    def test_check_ept_head_nok_no_head(self):
        # Given
        upload_dir = Path("tests/fixtures/ept/invalid_ept/no_head")

        # When
        result_status, result_ept_head = check_ept_head(
            self.WORK_DIR_PATH, upload_dir, self.EPT_SRS
        )

        # Then
        self.assertEqual(result_status, Status.FAILURE)
        self.assertEqual(result_ept_head, None)

    def test_check_ept_head_nok_multiple_heads(self):
        # Given
        upload_dir = Path("tests/fixtures/ept/invalid_ept/multiple_heads")

        # When
        result_status, result_ept_head = check_ept_head(
            self.WORK_DIR_PATH, upload_dir, self.EPT_SRS
        )

        # Then
        self.assertEqual(result_status, Status.FAILURE)
        self.assertEqual(result_ept_head, None)

    def test_get_extent_ok(self):
        # Given
        expected: GpfExtent = GpfExtent(
            extent=tuple[float](
                [
                    -105.19989506297283,
                    -105.21020072814241,
                    39.66929285047644,
                    39.66129150903086,
                ]
            )
        )

        # When
        result: GpfExtent = get_extent(Path(self.EPT_HEAD))

        # Then
        self.assertEqual(result, expected)

    def test_get_extent_ko(self):
        # When
        result: GpfExtent = get_extent(
            Path("tests/fixtures/ept/invalid_ept/bad_srs/ept.json")
        )

        # Then
        self.assertEqual(result, None)

    def test_get_srs(self):
        # Given
        ept_head = Path("tests/fixtures/ept/valid_ept/ept.json")
        expected_srs = "EPSG:26913"

        # When
        result = get_srs(ept_head)

        # Then
        self.assertEqual(result, expected_srs)

    def test_get_srs_nok_bad_head(self):
        # Given
        ept_head = Path("tests/fixtures/ept/invalid_ept/no_head/ept.json")

        # When
        result = get_srs(ept_head)

        # Then
        self.assertIsNone(result)

    def test_clean_pyramid_temporary_files(self):
        # Given
        source: str = "tests/fixtures/entwine/temp_files_ept"
        with patch("os.remove") as remove_mock:
            # When
            clean_pyramid_temporary_files(Path(source))

            # Then
            remove_mock.assert_has_calls(
                [
                    call(Path(f"{source}/ept-1.json")),
                    call(Path(f"{source}/ept-3.json")),
                    call(Path(f"{source}/ept-build-555.json")),
                    call(Path(f"{source}/ept-data/0-0-0-0-1.laz")),
                    call(Path(f"{source}/ept-data/0-2-3-1-10.laz")),
                    call(Path(f"{source}/ept-hierarchy/0-0-0-0-3.json")),
                    call(Path(f"{source}/ept-hierarchy/0-0-0-0-12.json")),
                    call(Path(f"{source}/ept-sources/manifest-7.json")),
                    call(Path(f"{source}/ept-sources/manifest-78.json")),
                ],
                any_order=True,
            )


# ################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()
