#! python3  # noqa E265
# standard
import unittest
from pathlib import Path
from unittest.mock import MagicMock

from gpf_entrepot_toolbelt.orchestrator.models.bbox import GpfBbox

# 3rd party
from gpf_entrepot_toolbelt.orchestrator.status import Status
from gpf_entrepot_toolbelt.utils.gpf_logger import gpf_logger_script

# project
import gpf_common_point_cloud
from gpf_common_point_cloud.utils.laslaz_utils import (
    READER_NAME,
    _is_file_in_tiling,
    check_laslaz_file_consistency,
    correct_max_coord,
    correct_min_coord,
    get_name_from_template,
    is_template_valid,
    pdal_info_from_las,
)

logger = gpf_logger_script(verbosity=0, title="TestLaslazUtils")


# ################################
# ########## Classes #############
# ################################


class TestLaslazUtils(unittest.TestCase):
    """Test paramaters abstraction class."""

    FILE: Path = Path("tests/fixtures/laslaz/tile_4860_67320.copc.laz")
    SRID: int = 2154
    SRS: str = "EPSG:" + str(SRID)
    TEST_GET_NAME_FROM_TEMPLATE_METADATA: dict = {
        READER_NAME: {
            "minx": 486000,
            "miny": 6731000,
            "maxx": 487000,
            "maxy": 6732000,
        }
    }
    BBOX: GpfBbox = GpfBbox([487000, 486000, 6732000, 6731000])

    def test_check_laslaz_file_consistency_ok_minimal(self):
        # Given
        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las = MagicMock(
            return_value={
                READER_NAME: {
                    "major_version": "X",
                    "minor_version": "X",
                    "srs": {"json": {"id": {"code": self.SRID}}},
                    "minx": 486000,
                    "miny": 6731000,
                    "maxx": 487000,
                    "maxy": 6732000,
                }
            }
        )

        # When
        status, bbox = check_laslaz_file_consistency(self.FILE)

        # Then
        self.assertEqual(status, Status.SUCCESS)
        self.assertEqual(bbox, self.BBOX)

        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las.assert_called_with(
            self.FILE
        )

    def test_check_laslaz_file_consistency_ok_complete(self):
        # Given
        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las = MagicMock(
            return_value={
                READER_NAME: {
                    "major_version": "FAKE",
                    "minor_version": "FAKE",
                    "srs": {"json": {"id": {"code": self.SRID}}},
                    "minx": 486000,
                    "miny": 6731000,
                    "maxx": 487000,
                    "maxy": 6732000,
                }
            }
        )

        # When
        status, bbox = check_laslaz_file_consistency(
            self.FILE,
            expected_srs=self.SRS,
            template="tile_{x_min/100}_{y_max/100}.copc.laz",
            origin={"x": -5000, "y": 5000},
            tile_size=1000,
        )

        # Then
        self.assertEqual(status, Status.SUCCESS)
        self.assertEqual(bbox, self.BBOX)

        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las.assert_called_with(
            self.FILE
        )

    def test_check_laslaz_file_consistency_ko_bad_pipeline(self):
        """Teste la fonction check_file_consistency"""
        # Given
        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las = MagicMock(
            return_value=None
        )

        # When
        status, bbox = check_laslaz_file_consistency(self.FILE)

        # Then
        self.assertEqual(status, Status.FAILURE)
        self.assertEqual(bbox, None)

        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las.assert_called_with(
            self.FILE
        )

    def test_check_laslaz_file_consistency_ko_bad_srs(self):
        # Given
        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las = MagicMock(
            return_value={
                READER_NAME: {
                    "major_version": "X",
                    "minor_version": "X",
                    "srs": {"json": {"id": {"code": self.SRID}}},
                    "minx": 486000,
                    "miny": 6731000,
                    "maxx": 487000,
                    "maxy": 6732000,
                }
            }
        )

        # When
        status, bbox = check_laslaz_file_consistency(
            self.FILE,
            expected_srs="EPSG:4326",
        )

        # Then
        self.assertEqual(status, Status.FAILURE)
        self.assertEqual(bbox, self.BBOX)

        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las.assert_called_with(
            self.FILE
        )

    def test_check_laslaz_file_consistency_ko_no_srs(self):
        # Given
        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las = MagicMock(
            return_value={
                READER_NAME: {
                    "major_version": "X",
                    "minor_version": "X",
                    "minx": 486000,
                    "miny": 6731000,
                    "maxx": 487000,
                    "maxy": 6732000,
                }
            }
        )

        # When
        status, bbox = check_laslaz_file_consistency(
            self.FILE,
            expected_srs="EPSG:4326",
        )

        # Then
        self.assertEqual(status, Status.SUCCESS)
        self.assertEqual(bbox, self.BBOX)

        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las.assert_called_with(
            self.FILE
        )

    def test_check_laslaz_file_consistency_ko_bad_tiling_origin(self):
        # Given
        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las = MagicMock(
            return_value={
                READER_NAME: {
                    "major_version": "X",
                    "minor_version": "X",
                    "srs": {"json": {"id": {"code": self.SRID}}},
                    "minx": 486500,
                    "miny": 6731200,
                    "maxx": 487500,
                    "maxy": 6732200,
                }
            }
        )

        # When
        status, bbox = check_laslaz_file_consistency(
            self.FILE,
            origin={"x": -5000, "y": 5000},
            tile_size=1000,
        )

        # Then
        self.assertEqual(status, Status.FAILURE)
        self.assertEqual(bbox, self.BBOX)

        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las.assert_called_with(
            self.FILE
        )

    def test_check_laslaz_file_consistency_ko_bad_tiling_size(self):
        # Given
        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las = MagicMock(
            return_value={
                READER_NAME: {
                    "major_version": "X",
                    "minor_version": "X",
                    "srs": {"json": {"id": {"code": self.SRID}}},
                    "minx": 486000,
                    "miny": 6731000,
                    "maxx": 487200,
                    "maxy": 6732200,
                }
            }
        )

        # When
        status, bbox = check_laslaz_file_consistency(
            self.FILE,
            origin={"x": -5000, "y": 5000},
            tile_size=1000,
        )

        # Then
        self.assertEqual(status, Status.FAILURE)
        self.assertEqual(bbox, self.BBOX)

        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las.assert_called_with(
            self.FILE
        )

    def test_check_laslaz_file_consistency_ko_bad_template(self):
        # Given
        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las = MagicMock(
            return_value={
                READER_NAME: {
                    "major_version": "FAKE",
                    "minor_version": "FAKE",
                    "srs": {"json": {"id": {"code": self.SRID}}},
                    "minx": 486000,
                    "miny": 6731000,
                    "maxx": 487000,
                    "maxy": 6732000,
                }
            }
        )

        # When
        status, bbox = check_laslaz_file_consistency(
            self.FILE,
            template="tile_{nope/100}_{y_max/100}.copc.laz",
            origin={"x": 0, "y": 0},
            tile_size=1000,
        )

        # Then
        self.assertEqual(status, Status.FAILURE)
        self.assertEqual(bbox, self.BBOX)

        gpf_common_point_cloud.utils.laslaz_utils.pdal_info_from_las.assert_called_with(
            self.FILE
        )

    def test_is_template_valid_ok_minimal(self):
        # When
        result: bool = is_template_valid("tile_{x_min}_{y_max}.copc.laz")

        # Then
        self.assertEqual(result, True)

    def test_is_template_valid_ok_maximal(self):
        # When
        result: bool = is_template_valid(
            "tile_{x_min}_{x_max}_{x_mean}_{y_max}.copc.laz"
        )

        # Then
        self.assertEqual(result, True)

    def test_is_template_valid_ko_no_field(self):
        # When
        result: bool = is_template_valid("tile.copc.laz")

        # Then
        self.assertEqual(result, False)

    def test_is_template_valid_ko_only_x(self):
        # When
        result: bool = is_template_valid("tile_{x_min}_{x_max}_{x_mean}.copc.laz")

        # Then
        self.assertEqual(result, False)

    def test_is_template_valid_ko_only_y(self):
        # When
        result: bool = is_template_valid("tile_{y_min}_{y_max}_{y_mean}.copc.laz")

        # Then
        self.assertEqual(result, False)

    def test_get_name_from_template_ok_no_field(self):
        # When
        result: str = get_name_from_template(
            self.TEST_GET_NAME_FROM_TEMPLATE_METADATA,
            template="tile.copc.laz",
            origin={"x": 0, "y": 0},
            tile_size=1000,
        )

        # Then
        self.assertEqual(result, "tile.copc.laz")

    def test_get_name_from_template_ok_no_modifier(self):
        # When
        result: str = get_name_from_template(
            self.TEST_GET_NAME_FROM_TEMPLATE_METADATA,
            template="tile_{x_min}_{y_max}.copc.laz",
            origin={"x": 0, "y": 0},
            tile_size=1000,
        )

        # Then
        self.assertEqual(result, "tile_486000_6732000.copc.laz")

    def test_get_name_from_template_ok_divider(self):
        # When
        result: str = get_name_from_template(
            self.TEST_GET_NAME_FROM_TEMPLATE_METADATA,
            template="tile_{x_min/100}_{y_max/1000}.copc.laz",
            origin={"x": 0, "y": 0},
            tile_size=1000,
        )

        # Then
        self.assertEqual(result, "tile_4860_6732.copc.laz")

    def test_get_name_from_template_ok_formatters(self):
        # When
        result: str = get_name_from_template(
            self.TEST_GET_NAME_FROM_TEMPLATE_METADATA,
            template="{x_min:3f}_{x_min:015f}_{x_min:.01f}_{x_min:015.03f}",
            origin={"x": 0, "y": 0},
            tile_size=1000,
        )

        # Then
        self.assertEqual(
            result, "486000.000000_00486000.000000_486000.0_00000486000.000"
        )

    def test_get_name_from_template_ko_bad_field(self):
        # When
        result: str = get_name_from_template(
            self.TEST_GET_NAME_FROM_TEMPLATE_METADATA,
            template="{bad}",
            origin={"x": 0, "y": 0},
            tile_size=1000,
        )

        # Then
        self.assertEqual(result, None)

    def test_get_name_from_template_no_conflict(self):
        # Given
        metadata1 = pdal_info_from_las(
            Path("./tests/fixtures/laslaz/tile_192_60508.copc.laz")
        )
        metadata2 = pdal_info_from_las(
            Path("./tests/fixtures/laslaz/tile_192_60509.copc.laz")
        )

        # When
        result1: str = get_name_from_template(
            metadata1,
            template="DALLE_{x_min/100:06.0f}_{y_max/100:06.0f}.copc.laz",
            origin={"x": 100, "y": 100},
            tile_size=100,
        )
        result2: str = get_name_from_template(
            metadata2,
            template="DALLE_{x_min/100:06.0f}_{y_max/100:06.0f}.copc.laz",
            origin={"x": 100, "y": 100},
            tile_size=100,
        )

        # Then
        self.assertEqual("DALLE_000193_060510.copc.laz", result1)
        self.assertEqual("DALLE_000193_060511.copc.laz", result2)

    def test_correct_max_coord_tile_size(self):
        # Given
        coord = 6051000.08
        origin = 100
        tile_size = 100

        # When
        result = correct_max_coord(coord, origin, tile_size)

        # Then
        self.assertEqual(6051100, result)

    def test_correct_min_coord_tile_size(self):
        # Given
        coord = 19089.61
        origin = 100
        tile_size = 100

        # When
        result = correct_min_coord(coord, origin, tile_size)

        # Then
        self.assertEqual(19000.0, result)

    def test_correct_max_coord_tile_size_le_0(self):
        # Given
        coord = 486930
        origin = 0
        tile_size = 0

        # When
        result = correct_max_coord(coord, origin, tile_size)

        # Then
        self.assertEqual(coord, result)

    def test_correct_min_coord_tile_size_le_0(self):
        # Given
        coord = 486930
        origin = 0
        tile_size = 0

        # When
        result = correct_min_coord(coord, origin, tile_size)

        # Then
        self.assertEqual(coord, result)

    def test_is_file_in_tiling(self):
        # Given
        test_is_file_in_tiling_ko_x: dict = {
            READER_NAME: {
                "minx": 486000,
                "miny": 6731000,
                "maxx": 487200,
                "maxy": 6732000,
            }
        }
        test_is_file_in_tiling_ko_y: dict = {
            READER_NAME: {
                "minx": 486000,
                "miny": 6730999.99,
                "maxx": 487000,
                "maxy": 6732000,
            }
        }

        # When
        result_success = _is_file_in_tiling(
            self.TEST_GET_NAME_FROM_TEMPLATE_METADATA,
            origin={"x": 0, "y": 0},
            tile_size=1000,
        )
        result_ko_x = _is_file_in_tiling(
            test_is_file_in_tiling_ko_x, origin={"x": 0, "y": 0}, tile_size=1000
        )
        result_ko_y = _is_file_in_tiling(
            test_is_file_in_tiling_ko_y, origin={"x": 0, "y": 0}, tile_size=1000
        )

        # Then
        self.assertEqual(True, result_success)
        self.assertEqual(False, result_ko_x)
        self.assertEqual(False, result_ko_y)


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()
