#! python3  # noqa E265

# standard
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from gpf_entrepot_toolbelt.orchestrator.status import Status

# 3rd party
from gpf_entrepot_toolbelt.utils.gpf_logger import gpf_logger_script

from gpf_common_point_cloud.models import PdalTileParameters

# project
from gpf_common_point_cloud.utils.laslaz_utils import (
    MAXX_FIELD,
    MAXY_FIELD,
    MINX_FIELD,
    MINY_FIELD,
    READER_NAME,
)
from gpf_common_point_cloud.utils.pdal_utils import (
    pdal_convert_las_to_copc,
    pdal_info_from_ept,
    pdal_info_from_las,
    pdal_merge,
    pdal_tile_las,
)

logger = gpf_logger_script(verbosity=0, title="TestPdalUtils")


# ################################
# ########## Classes #############
# ################################


class TestPdalUtils(unittest.TestCase):
    """Test paramaters abstraction class."""

    def test_pdal_info_from_las_ok(self):
        # Given
        file: Path = Path("tests/fixtures/laslaz/tile_4860_67320.copc.laz")

        # When
        result: dict = pdal_info_from_las(file)

        # Then
        self.assertEqual(result[READER_NAME]["copc"], True)
        self.assertEqual(result[READER_NAME]["compressed"], True)
        self.assertEqual(result[READER_NAME]["count"], 262453)
        self.assertEqual(result[READER_NAME][MINX_FIELD], 486900)
        self.assertEqual(result[READER_NAME][MINY_FIELD], 6731400)
        self.assertEqual(result[READER_NAME]["minz"], 34.42)
        self.assertEqual(result[READER_NAME][MAXX_FIELD], 486999.99)
        self.assertEqual(result[READER_NAME][MAXY_FIELD], 6731499.99)
        self.assertEqual(
            result[READER_NAME]["maxz"],
            64.28,
        )

    def test_pdal_info_from_las_ok_with_bad_extension(self):
        # Given
        file: Path = Path("tests/fixtures/laslaz/tile_4869_67314.copc.txt")

        # When
        result: dict = pdal_info_from_las(file)

        # Then
        self.assertTrue(READER_NAME in result and result[READER_NAME])

    def test_pdal_info_from_las_ko_bad_laz(self):
        # Given
        file: Path = Path("tests/fixtures/laslaz/tile_4869_67314.copc.KO.laz")

        # When
        result: dict = pdal_info_from_las(file)

        # Then
        self.assertEqual(result, None)

    def test_pdal_info_from_las_ok_terrascan_warning(self):
        # Given
        file: Path = Path("tests/fixtures/laslaz/laz_file_terrascan_KO.laz")

        # When
        result: dict = pdal_info_from_las(file)

        # Then
        self.assertEqual(result[READER_NAME]["copc"], False)
        self.assertEqual(result[READER_NAME]["compressed"], True)
        self.assertEqual(result[READER_NAME]["count"], 473834)
        self.assertEqual(result[READER_NAME][MINX_FIELD], 924450)
        self.assertEqual(result[READER_NAME][MINY_FIELD], 6288450)
        self.assertEqual(result[READER_NAME]["minz"], 347.33)
        self.assertEqual(result[READER_NAME][MAXX_FIELD], 924550)
        self.assertEqual(result[READER_NAME][MAXY_FIELD], 6288550)
        self.assertEqual(result[READER_NAME]["maxz"], 381.9)

    @patch("subprocess.run")
    def test_merge_tiles(self, mock_run):
        # Given
        mock_proc = MagicMock(returncode=0)
        mock_proc.communicate.return_value = (b"standard output", b"")
        mock_run.return_value = mock_proc

        tiles_to_merge = [Path("/fake/path/tile1.laz"), Path("/fake/path/tile2.laz")]
        output_file_path = Path("/fake/output/merged.laz")

        # When
        result = pdal_merge(tiles_to_merge, output_file_path)

        # Then
        files_to_process = " ".join([str(tile) for tile in tiles_to_merge])
        command_to_run = f"pdal merge --verbose=2 {files_to_process} {output_file_path}"

        mock_run.assert_called_once_with(
            command_to_run, shell=True, capture_output=True, text=True
        )
        self.assertEqual(Status.SUCCESS, result)

    def test_merge_tiles_no_files(self):
        # Given
        tiles_to_merge = []
        output_file_path = Path("/fake/output/merged.laz")

        # When
        result = pdal_merge(tiles_to_merge, output_file_path)

        # Then
        self.assertEqual(Status.SUCCESS, result)

    @patch("subprocess.run")
    def test_merge_tiles_error(self, mock_run):
        # Given
        mock_proc = MagicMock(returncode=3)
        mock_proc.communicate.return_value = (b"standard output", b"standard error")
        mock_run.return_value = mock_proc

        tiles_to_merge = [Path("/fake/path/tile1.laz"), Path("/fake/path/tile2.laz")]
        output_file_path = Path("/fake/output/merged.laz")

        # When
        result = pdal_merge(tiles_to_merge, output_file_path)

        # Then
        files_to_process = " ".join([str(tile) for tile in tiles_to_merge])
        command_to_run = f"pdal merge --verbose=2 {files_to_process} {output_file_path}"

        mock_run.assert_called_once_with(
            command_to_run, shell=True, capture_output=True, text=True
        )
        self.assertEqual(Status.FAILURE, result)

    @patch("subprocess.run")
    def test_merge_tiles_exception(self, mock_run):
        # Given
        mock_run.side_effect = Exception("Test exception")
        tiles_to_merge = [Path("/fake/path/tile1.laz"), Path("/fake/path/tile2.laz")]
        output_file_path = Path("/fake/output/merged.laz")

        # When
        result = pdal_merge(tiles_to_merge, output_file_path)

        # Then
        files_to_process = " ".join([str(tile) for tile in tiles_to_merge])
        command_to_run = f"pdal merge --verbose=2 {files_to_process} {output_file_path}"

        mock_run.assert_called_once_with(
            command_to_run, shell=True, capture_output=True, text=True
        )
        self.assertEqual(Status.FAILURE, result)

    @patch("subprocess.run")
    def test_pdal_tile_las(self, mock_run):
        # Given
        pdal_tile_params = PdalTileParameters(
            out_srs="EPSG:4326", origin_x=0.0, origin_y=0.0, tile_size=1000
        )
        input_dir = Path("/path/to/input")
        output_dir = Path("/path/to/output")
        mock_proc = MagicMock(returncode=0)
        mock_proc.communicate.return_value = (b"standard output", b"")
        mock_run.return_value = mock_proc

        # When
        result = pdal_tile_las(pdal_tile_params, input_dir, output_dir)

        # Then
        in_pattern = f'"{input_dir}/*"'
        command = [
            "pdal",
            "tile",
            "--writers.las.forward",
            "header",
            "--writers.las.minor_version",
            "4",
            "-i",
            in_pattern,
            "-o",
            f'"{output_dir}/tile_#.laz"',
        ]

        command.extend([f'--out_srs="{pdal_tile_params.out_srs}"'])
        command.extend([f'--writers.las.a_srs="{pdal_tile_params.out_srs}"'])
        command.extend(["--origin_x", str(pdal_tile_params.origin_x)])
        command.extend(["--origin_y", str(pdal_tile_params.origin_y)])
        command.extend(["--length", str(pdal_tile_params.tile_size)])
        command = " ".join(command)

        mock_run.assert_called_once_with(
            command, shell=True, capture_output=True, text=True
        )
        self.assertEqual(Status.SUCCESS, result)

    @patch("subprocess.run")
    def test_pdal_tile_las_error(self, mock_run):
        # Given
        pdal_tile_params = PdalTileParameters(
            out_srs="EPSG:4326", origin_x=0.0, origin_y=0.0, tile_size=1000
        )
        input_dir = Path("/path/to/input")
        output_dir = Path("/path/to/output")
        mock_proc = MagicMock(returncode=3)
        mock_proc.communicate.return_value = (b"standard output", b"standard error")
        mock_run.return_value = mock_proc

        # When
        result = pdal_tile_las(pdal_tile_params, input_dir, output_dir)

        # Then
        in_pattern = f'"{input_dir}/*"'
        command = [
            "pdal",
            "tile",
            "--writers.las.forward",
            "header",
            "--writers.las.minor_version",
            "4",
            "-i",
            in_pattern,
            "-o",
            f'"{output_dir}/tile_#.laz"',
        ]

        command.extend([f'--out_srs="{pdal_tile_params.out_srs}"'])
        command.extend([f'--writers.las.a_srs="{pdal_tile_params.out_srs}"'])
        command.extend(["--origin_x", str(pdal_tile_params.origin_x)])
        command.extend(["--origin_y", str(pdal_tile_params.origin_y)])
        command.extend(["--length", str(pdal_tile_params.tile_size)])
        command = " ".join(command)
        mock_run.assert_called_once_with(
            command, shell=True, capture_output=True, text=True
        )
        self.assertEqual(Status.FAILURE, result)

    @patch("subprocess.run")
    def test_pdal_tile_las_exception(self, mock_run):
        # Given
        pdal_tile_params = PdalTileParameters(
            out_srs="EPSG:4326", origin_x=0.0, origin_y=0.0, tile_size=1000
        )
        input_dir = Path("/path/to/input")
        output_dir = Path("/path/to/output")
        mock_run.side_effect = Exception("Test exception")

        # When
        result = pdal_tile_las(pdal_tile_params, input_dir, output_dir)

        # Then
        in_pattern = f'"{input_dir}/*"'
        command = [
            "pdal",
            "tile",
            "--writers.las.forward",
            "header",
            "--writers.las.minor_version",
            "4",
            "-i",
            in_pattern,
            "-o",
            f'"{output_dir}/tile_#.laz"',
        ]

        command.extend([f'--out_srs="{pdal_tile_params.out_srs}"'])
        command.extend([f'--writers.las.a_srs="{pdal_tile_params.out_srs}"'])
        command.extend(["--origin_x", str(pdal_tile_params.origin_x)])
        command.extend(["--origin_y", str(pdal_tile_params.origin_y)])
        command.extend(["--length", str(pdal_tile_params.tile_size)])
        command = " ".join(command)
        mock_run.assert_called_once_with(
            command, shell=True, capture_output=True, text=True
        )
        self.assertEqual(Status.FAILURE, result)

    def test_pdal_convert_las_to_copc(self):
        # Given
        input_file_path = Path("./tests/fixtures/laslaz/tile_4860_67320.copc.laz")
        output_file_path = Path("./tests/fixtures/laslaz/converted.copc.laz")

        # When
        result = pdal_convert_las_to_copc(input_file_path, output_file_path)

        # Then
        self.assertEqual(Status.SUCCESS, result)
        self.assertTrue(output_file_path.exists())
        output_file_path.unlink()

    @patch("pdal.Reader.las")
    def test_pdal_convert_las_to_copc_exception(self, mock_pdal_reader):
        # Given
        input_file_path = Path("./tests/fixtures/laslaz/tile_4860_67320.copc.laz")
        output_file_path = Path("./tests/fixtures/laslaz/converted.copc.laz")
        mock_pdal_reader.side_effect = Exception("Test exception")

        # When
        result = pdal_convert_las_to_copc(input_file_path, output_file_path)

        # Then
        self.assertEqual(Status.FAILURE, result)

    def test_pdal_info_from_ept_success(self):
        # Given # When
        result = pdal_info_from_ept(Path("./tests/fixtures/ept/valid_ept/ept.json"))

        # Then
        self.assertEqual(Status.SUCCESS, result)

    def test_pdal_info_from_ept_failure(self):
        # Given # When
        result = pdal_info_from_ept(
            Path("./tests/fixtures/ept/invalid_ept/missing_files/ept.json")
        )

        # Then
        self.assertEqual(Status.FAILURE, result)

    @patch("subprocess.run")
    def test_pdal_info_from_ept_exception(self, mock_run):
        # Given
        input_file = Path("/path/to/input/file.json")
        mock_run.side_effect = Exception("Test exception")

        # When
        result = pdal_info_from_ept(input_file)

        # Then
        command = (
            f"pdal info --verbose=2 --readers.ept.ignore_unreadable true {input_file}"
        )
        mock_run.assert_called_once_with(
            command, shell=True, capture_output=True, text=True
        )
        self.assertEqual(Status.TECHNICAL_ERROR, result)


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()
